"""
Implementation of WS Server.

WebSocketServer:
    Create WS Server.
"""

from typing import Any, Dict

from ros2bridge.server.register_client import register_client

from tornado.websocket import WebSocketHandler


class WebSocketServer(WebSocketHandler):
    """
    Create new websocket server.

    Create a websocket server using tornado.
    Handles all incoming and outgoing clients and requests.

    Attributes
    ----------
    connected_clients: Dict[object, Any]
        List of client connected to the server.

    Methods
    -------
    open(self) -> None:
        Handle new client connection.

    on_message(self, message: str) -> None:
        Message received from client.
    on_close(self) -> None:
        Client disconnected.

    send_message(self, message: str) -> None
        Send given message to client.

    check_origin(self, _: str) -> bool:
        Allow CROS.

    set_default_headers(self) -> None:
        CROS Headers.
    """

    connected_clients: Dict[object, Any] = {}

    def open(self) -> None:  # type: ignore [override] # noqa: A003
        """Handle new client connection."""
        cls = self.__class__

        # Register new client
        _client = register_client(
            socket=cls,  # type: ignore [arg-type]
            client=self  # type: ignore [arg-type]
        )
        _client_name = list(_client.keys())[0]
        print(f'Client connected: {_client_name}')

        # Update client list
        cls.connected_clients[_client_name] = _client[_client_name]

    def on_message(self, message: str) -> None:  # type: ignore [override]
        """Message received from client.

        Args:
            message (str): Message received from client.
        """
        ...

    def on_close(self) -> None:
        """Client disconnected."""
        ...

    def send_message(self, message: str) -> None:
        """Send given message to client.

        Args:
            message (str): Message to send.
        """
        ...

    def check_origin(self, _: str) -> bool:
        """Allow CROS.

        Args:
            _ (str): Origin

        Returns:
            bool: True, Accept all origin.
        """
        return True

    def set_default_headers(self) -> None:
        """CROS Headers."""
        self.set_header('Content-Type', 'application/json')
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', 'x-requested-with')
        super().set_default_headers()



"""
Blueprint of WS Server.

WSServer:
    WS Server.
"""

from typing import Protocol


class WSServer(Protocol):
    """
    Create new websocket server.

    Create a websocket server using tornado.
    Handles all incoming and outgoing clients and requests.

    Attributes
    ----------
    connected_clients: Dict[object, Any]
        List of client connected to the server.

    Methods
    -------
    on_open(self) -> None:
        Handle new client connection.

    on_message(self, message: str) -> None:
        Message received from client.
    on_close(self) -> None:
        Client disconnected.

    send_message(self, message: str) -> None
        Send given message to client.

    check_origin(self, _: str) -> bool:
        Allow CROS.

    set_default_headers(self) -> None:
        CROS Headers.
    """

    def on_open(self) -> None:
        """Handle new client connection."""
        ...

    def on_message(self, message: str) -> None:
        """Message received from client.

        Args:
            message (str): Message received from client.
        """
        ...

    def on_close(self) -> None:
        """Client disconnected."""
        ...

    def send_message(self, message: str) -> None:
        """Send given message to client.

        Args:
            message (str): Message to send.
        """
        ...

    def check_origin(self, _: str) -> bool:
        """Allow CROS.

        Args:
            _ (str): Origin

        Returns:
            bool: True, Accept all origin.
        """
        ...

    def set_default_headers(self) -> None:
        """CROS Headers."""
        ...



"""Test WS Connection."""


from tests.utils.pre_post import PrePost

from websocket import WebSocket


pre_post = PrePost()


def test_connection() -> None:
    """Test websocket connection."""
    _ws_url = pre_post.get_ws_url()

    _ws = WebSocket()
    _ws.connect(_ws_url)

    assert _ws.connected



"""
Registering new client.

register_client: Register new WS client
"""

from secrets import token_hex
from typing import Any, Dict, Type

from rclpy.node import Node

from ros2bridge.protocols.ws_server import WSServer as WS


def register_client(socket: Type[WS], client: WS) -> Dict[str, Any]:
    """
    Register new ws client.

    Args:
        socket (Type[WS]): WS server.
        client (WS): Connected client.

    Returns:
        Dict[str, Any]: Client info.
    """
    _client_id = f'client_{token_hex(8)}'
    _node = Node(_client_id)

    _operations = {
        'publish': None,
        'subscribe': None,
        'srv_client': None,
        'action_client': None
    }

    return {
        _client_id: {
            'client': client,
            'client_id': _client_id,
            'client_node': _node,
            'client_socket': socket,

            'operations': _operations,
            'publisher': {},
            'subscriber': {},
            'srv_client': {},
            'action_client': {},
            'terminate': []
        }
    }
